from .tile import Tile
from .player import PlayerType
from .shape import BoardShape
from .path import PathPair
from typing import Optional, Union


class Piece:
    """
    A piece on a board.
    """
    __slots__ = ("_owner", "_path_index")

    _owner: PlayerType
    _path_index: int

    def __init__(self, owner: PlayerType, path_index: int):
        """
        Initializes a new Piece instance with the given owner and path index.

        Parameters:
            owner (PlayerType): The player that owns this piece.
            path_index (int): The index of the piece on its owner's path. Must be non-negative.

        Raises:
            ValueError: If the provided path_index is negative.
        """
        if path_index < 0:
            raise ValueError(f"The path index cannot be negative: {path_index}")

        self._owner = owner
        self._path_index = path_index

    @property
    def owner(self) -> PlayerType:
        """
        The player that owns this piece.
        """
        return self._owner

    @property
    def path_index(self) -> int:
        """
        The index of the piece on its owner player's path.
        """
        return self._path_index

    def __hash__(self) -> int:
        return hash((self._owner, self._path_index))

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False

        return self._owner == other._owner \
            and self._path_index == other._path_index

    @staticmethod
    def to_char(piece: Optional['Piece']) -> str:
        """
        Converts the given piece to a single character that can be used
        to textually represent the owner of a piece.
        """
        return PlayerType.to_char(piece.owner if piece is not None else None)


class Board:
    """
    Stores the placement of pieces on the tiles of a Royal Game of Ur board.
    """
    __slots__ = ("_shape", "_width", "_height", "_pieces")

    _shape: BoardShape
    _width: int
    _height: int
    _pieces: list[Optional[Piece]]

    def __init__(
            self, board_or_shape: Union[BoardShape, 'Board']
    ):
        if isinstance(board_or_shape, BoardShape):
            shape = board_or_shape
        else:
            shape = board_or_shape.shape

        self._shape = shape
        self._width = shape.width
        self._height = shape.height

        if isinstance(board_or_shape, Board):
            self._pieces = [*board_or_shape._pieces]
        else:
            self._pieces = [None for _ in range(self._width * self._height)]

    def copy(self) -> 'Board':
        """
        Creates an exact copy of this board.
        """
        return Board(self)

    def _calc_tile_index(self, ix: int, iy: int) -> int:
        """
        We use a 1d array for storing pieces. This converts
        from 2d indices into an index to access that 1d array.
        """
        if ix < 0 or iy < 0 or ix >= self._width or iy >= self._height:
            raise ValueError(f"There is no tile at the indices ({ix}, {iy})")

        return iy * self._width + ix

    @property
    def shape(self) -> BoardShape:
        """
        The shape of this board.
        """
        return self._shape

    @property
    def width(self) -> int:
        """
        The width of this board.
        """
        return self._width

    @property
    def height(self) -> int:
        """
        The height of this board.
        """
        return self._height

    def contains(self, tile: Tile) -> bool:
        """
        Determines whether the given tile falls within the bounds of this board.
        """
        return self._shape.contains(tile)

    def contains_indices(self, ix: int, iy: int) -> bool:
        """
        Determines whether the tile at the indices (ix, iy),
        0-based, falls within the bounds of this board.
        """
        return self._shape.contains_indices(ix, iy)

    def get(self, tile: Tile) -> Optional[Piece]:
        """
        Gets the piece on the given tile. Returns None if there
        is no piece on the tile.
        """
        if not self._shape.contains(tile):
            raise ValueError(f"There is no tile at {tile}")

        index = self._calc_tile_index(tile.ix, tile.iy)
        return self._pieces[index]

    def get_by_indices(self, ix: int, iy: int) -> Optional[Piece]:
        """
        Gets the piece on the tile at the indices (ix, iy), 0-based.
        Returns None if there is no piece on the tile.
        """
        if not self._shape.contains_indices(ix, iy):
            raise ValueError(f"There is no tile at the indices ({ix}, {iy})")

        index = self._calc_tile_index(ix, iy)
        return self._pieces[index]

    def set(self, tile: Tile, piece: Optional[Piece]) -> Optional[Piece]:
        """
        Sets the piece on the given tile to the provided piece. If the provided
        piece is None, it removes any piece on the tile. Returns the piece
        that was previously on the tile, or None if there was no
        piece on the tile.
        """
        if not self._shape.contains(tile):
            raise ValueError(f"There is no tile at {tile}")

        index = self._calc_tile_index(tile.ix, tile.iy)
        previous = self._pieces[index]
        self._pieces[index] = piece
        return previous

    def set_by_indices(self, ix: int, iy: int, piece: Optional[Piece]) -> Optional[Piece]:
        """
        Sets the piece on the tile at the indices (ix, iy), 0-based,
        to the provided piece. If the provided piece is None, it removes
        any piece on the tile. Returns the piece that was previously
        on the tile, or None if there was no piece on the tile.
        """
        if not self._shape.contains_indices(ix, iy):
            raise ValueError(f"There is no tile at the indices ({ix}, {iy})")

        index = self._calc_tile_index(ix, iy)
        previous = self._pieces[index]
        self._pieces[index] = piece
        return previous

    def clear(self):
        """
        Removes all pieces from this board.
        """
        for index in range(len(self._pieces)):
            self._pieces[index] = None

    def count_pieces(self, player: PlayerType) -> int:
        """
        Counts the number of pieces that are on the board for the given player.
        """
        piece_count = 0
        for piece in self._pieces:
            if piece is not None and piece.owner == player:
                piece_count += 1

        return piece_count

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False

        return self._shape == other._shape and self._pieces == other._pieces

    def to_string(self, column_delimiter: str = "\n", include_off_board_tiles: bool = True):
        """
        Writes the contents of this board into a String, where each column is separated by a delimiter.
        """
        builder = []
        for ix in range(self._width):
            if ix > 0:
                builder.append(column_delimiter)

            for iy in range(self._height):
                if self._shape.contains_indices(ix, iy):
                    piece = self.get_by_indices(ix, iy)
                    builder.append(Piece.to_char(piece))

                elif include_off_board_tiles:
                    builder.append(" ")

        return "".join(builder)

    def __repr__(self) -> str:
        return self.to_string()


class Move:
    """
    A move that can be made on a board.
    """
    __slots__ = (
        "_player", "_source", "_source_piece",
        "_dest", "_dest_piece", "_captured_piece"
    )

    _player: PlayerType
    _source: Optional[Tile]
    _source_piece: Optional[Piece]
    _dest: Optional[Tile]
    _dest_piece: Optional[Piece]
    _captured_piece: Optional[Piece]

    def __init__(
            self,
            player: PlayerType,
            source: Optional[Tile],
            source_piece: Optional[Piece],
            dest: Optional[Tile],
            dest_piece: Optional[Piece],
            captured_piece: Optional[Piece],
    ):
        if (source is None) != (source_piece is None):
            raise ValueError("source and source_piece must either be both null, or both non-null")
        if (dest is None) != (dest_piece is None):
            raise ValueError("dest and dest_piece must either be both null, or both non-null")
        if dest is None and captured_piece is not None:
            raise ValueError("Moves without a destination cannot have captured a piece")

        self._player = player
        self._source = source
        self._source_piece = source_piece
        self._dest = dest
        self._dest_piece = dest_piece
        self._captured_piece = captured_piece

    @property
    def player(self) -> PlayerType:
        """
        The instigator of this move.
        """
        return self._player

    @property
    def source(self) -> Optional[Tile]:
        """
        The origin of the move. If this is None, it represents
        that this move is introducing a new piece to the board.
        """
        return self._source

    @property
    def source_piece(self) -> Optional[Piece]:
        """
        The piece on the board to be moved, or None if this move
        is introducing a new piece to the board.
        """
        return self._source_piece

    @property
    def dest(self) -> Optional[Tile]:
        """
        The destination of the move. If this is None, it represents
        that this move is scoring a piece.
        """
        return self._dest

    @property
    def dest_piece(self) -> Optional[Piece]:
        """
        The piece to be placed at the destination, or None if
        this move is scoring a piece.
        """
        return self._dest_piece

    @property
    def captured_piece(self) -> Optional[Piece]:
        """
        The piece that will be captured by this move, or None if
        this move does not capture a piece.
        """
        return self._captured_piece

    def has_source(self) -> bool:
        """
        Determines whether this move is moving a piece on the board.
        """
        return self._source is not None

    def is_introducing_piece(self) -> bool:
        """
        Determines whether this move is moving a new piece onto the board.
        """
        return self._source is None

    def has_dest(self) -> bool:
        """
        Determines whether this moves a piece to a destination on the board.
        """
        return self._dest is not None

    def is_scoring_piece(self) -> bool:
        """
        Determines whether this move is moving a piece off of the board.
        """
        return self._dest is None

    def is_capture(self) -> bool:
        """
        Determines whether this move is capturing an existing piece on the board.
        """
        return self._captured_piece is not None

    def is_dest_rosette(self, shape: BoardShape) -> bool:
        """
        Determines whether this move will land a piece on a rosette. Under common
        rule sets, this will give another turn to the player.
        """
        return self._dest is not None and shape.is_rosette(self._dest)

    def get_source(self, paths: PathPair | None = None) -> Tile:
        """
        Gets the source piece of this move. If there is no source piece, in the
        case where a new piece is moved onto the board, this will throw an error.
        """
        if self._source is None:
            if paths is not None:
                return paths.get_start(self._player)

            raise RuntimeError("This move has no source, as it is introducing a piece")

        return self._source

    def get_source_piece(self) -> Piece:
        """
        Gets the source piece of this move. If there is no source piece, in the
        case where a new piece is moved onto the board, this will throw an error.
        """
        if self._source_piece is None:
            raise RuntimeError("This move has no source, as it is introducing a piece")

        return self._source_piece

    def get_dest(self, paths: PathPair | None = None) -> Tile:
        """
        Gets the destination tile of this move. If there is no destination tile,
        in the case where a piece is moved off the board, this will throw an error.
        """
        if self._dest is None:
            if paths is not None:
                return paths.get_end(self._player)

            raise RuntimeError("This move has no destination, as it is scoring a piece")

        return self._dest

    def get_dest_piece(self) -> Piece:
        """
        Gets the destination piece of this move. If there is no destination piece,
        in the case where a piece is moved off the board, this will throw an error.
        """
        if self._dest_piece is None:
            raise RuntimeError("This move has no destination, as it is scoring a piece")

        return self._dest_piece

    def get_captured_piece(self) -> Piece:
        """
        Gets the piece that will be captured by this move. If there is no piece
        that will be captured, this will throw an error.
        """
        if self._captured_piece is None:
            raise RuntimeError("This move does not capture a piece");

        return self._captured_piece

    def apply(self, board: Board):
        """
        Apply this move to update the given board.
        """
        if self.source is not None:
            board.set(self.source, None)
        if self.dest is not None:
            board.set(self.dest, self.dest_piece)

    def describe(self) -> str:
        """
        Generates an English description of this move.
        """
        scoring = self.is_scoring_piece()
        introducing = self.is_introducing_piece()

        if scoring and introducing:
            return "Introduce and score a piece"

        if scoring:
            return f"Score a piece from {self.get_source()}"

        builder = []
        if introducing:
            builder.append("Introduce a piece to ")
        else:
            builder.append(f"Move {self.get_source()} to ")

        if self.is_capture():
            builder.append("capture ")

        builder.append(str(self.get_dest()))
        return "".join(builder)

    def __hash__(self) -> int:
        return hash((
            self._source, self._source_piece,
            self._dest, self._dest_piece,
            self._captured_piece
        ))

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False

        return self._source == other._source and self._source_piece == other._source_piece \
            and self._dest == other._dest and self._dest_piece == other._dest_piece \
            and self._captured_piece == other._captured_piece



from royalur.model import (
    Piece, PlayerType, PlayerState, BoardShape, PathPair, Dice,
    Board, Roll, Move, GameSettings, GameMetadata
)
from .state import GameState, WaitingForRollGameState, WaitingForMoveGameState
from abc import ABC, abstractmethod
from typing import Callable


class PieceProvider(ABC):
    """
    An interface that provides instances of pieces. This may be used
    to store custom information with each piece, for situations such
    as adding stacking or unique piece behavior.
    """
    __slots__ = ()

    @abstractmethod
    def create_introduced(self, owner: PlayerType, new_path_index: int) -> Piece:
        """
        Generates a new piece to be introduced to the board.
        """
        pass

    @abstractmethod
    def create_moved(self, origin_piece: Piece, new_path_index: int) -> Piece:
        """
        Generates a piece that has been moved from another tile on the board.
        """
        pass


class PlayerStateProvider(ABC):
    """
    An interface that provides the manipulation of PlayerStates as a game progresses.
    """
    __slots__ = ()

    @abstractmethod
    def get_starting_piece_count(self) -> int:
        """
        Gets the number of pieces that players start with.
        """
        pass

    @abstractmethod
    def create(self, player: PlayerType):
        """
        Generates the starting state for the given player.
        """
        pass

    @abstractmethod
    def apply_piece_introduced(self, player_state: PlayerState, piece: Piece):
        """
        Generates a new player state that is a copy of the given player state,
        with the given piece introduced to the board.
        """
        pass

    @abstractmethod
    def apply_piece_captured(self, player_state: PlayerState, piece: Piece):
        """
        Generates a new player state that is a copy of the given player state,
        with the given piece captured.
        """
        pass

    @abstractmethod
    def apply_piece_scored(self, player_state: PlayerState, piece: Piece):
        """
        Generates a new player state that is a copy of the given player state,
        with the given piece scored.
        """
        pass


class RuleSet(ABC):
    """
    A set of rules that govern the play of a game of the Royal Game of Ur.
    """
    __slots__ = ("_board_shape", "_paths", "_dice_factory", "_piece_provider", "_player_state_provider")

    _board_shape: BoardShape
    _paths: PathPair
    _dice_factory: Callable[[], Dice]
    _piece_provider: PieceProvider
    _player_state_provider: PlayerStateProvider

    def __init__(
            self,
            board_shape: BoardShape,
            paths: PathPair,
            dice_factory: Callable[[], Dice],
            piece_provider: PieceProvider,
            player_state_provider: PlayerStateProvider
    ):
        self._board_shape = board_shape
        self._paths = paths
        self._dice_factory = dice_factory
        self._piece_provider = piece_provider
        self._player_state_provider = player_state_provider

    @property
    def settings(self) -> GameSettings:
        """
        The settings used for this rule set.
        """
        return GameSettings(
            self._board_shape,
            self._paths,
            self._dice_factory,
            self._player_state_provider.get_starting_piece_count(),
            self.are_rosettes_safe(),
            self.do_rosettes_grant_extra_rolls(),
            self.do_captures_grant_extra_rolls()
        )

    @property
    def board_shape(self) -> BoardShape:
        """
        The shape of the game board.
        """
        return self._board_shape

    @property
    def paths(self) -> PathPair:
        """
        The paths that each player must take around the board.
        """
        self._paths

    @property
    def dice_factory(self) -> Callable[[], Dice]:
        """
        The generator of dice that are used to generate dice rolls.
        """
        return self._dice_factory

    @property
    def piece_provider(self) -> PieceProvider:
        """
        Provides the manipulation of piece values.
        """
        return self._piece_provider

    @property
    def player_state_provider(self) -> PlayerStateProvider:
        """
        Provides the manipulation of player state values.
        """
        return self._player_state_provider

    @abstractmethod
    def are_rosettes_safe(self) -> bool:
        """
        Gets whether rosettes are considered safe squares in this rule set.
        """
        pass

    @abstractmethod
    def do_rosettes_grant_extra_rolls(self) -> bool:
        """
        Gets whether landing on rosette tiles grants an additional roll.
        """
        pass

    @abstractmethod
    def do_captures_grant_extra_rolls(self) -> bool:
        """
        Gets whether capturing a piece grants an additional roll.
        """
        pass

    @abstractmethod
    def generate_initial_game_state(self) -> GameState:
        """
        Generates the initial state for a game.
        """
        pass

    @abstractmethod
    def find_available_moves(
            self,
            board: Board,
            player: PlayerState,
            roll: Roll,
    ) -> list[Move]:
        """
        Finds all available moves from the given state.
        """
        pass

    @abstractmethod
    def apply_roll(
        self,
        state: WaitingForRollGameState,
        roll: Roll,
    ) -> list[GameState]:
        """
        Applies the given roll to the given state to generate the new state of the
        game. Multiple game states will be returned to include action game
        states for maintaining history. However, the latest or highest-index
        game state will represent the state of the game after the roll was made.
        """
        pass

    @abstractmethod
    def apply_move(
        self,
        state: WaitingForMoveGameState,
        move: Move,
    ) -> list[GameState]:
        """
        Applies the given move to the given state to generate the new state of
        the game. Multiple game states may be returned to include action game
        states for maintaining history. However, the latest or highest-index
        game state will represent the state of the game after the move was made.

        This method does not check that the given move is valid.
        """
        pass


class RuleSetProvider(ABC):
    """
    Creates rule sets to match game settings.
    """
    __slots__ = ()

    @abstractmethod
    def create(
            self,
            settings: GameSettings,
            metadata: GameMetadata,
    ) -> RuleSet:
        """
        Creates a rule set to match the given settings and game metadata.
        """
        pass



"""
This package contains the RoyalUr library.
"""

from .model import (
    Tile,
    PlayerType, PlayerState,
    PathPair, PathType,
    BoardShape, BoardType,
    Piece, Move, Board,
    Dice, DiceType,
    GameSettings,
    GameMetadata,
)
from .rules import (
    RuleSet, RuleSetProvider,
    SimpleRuleSet, SimpleRuleSetProvider,
)
from .game import (
    Game, GameBuilder,
)



import unittest
import random
from royalur import Game


class TestExample(unittest.TestCase):

    def test_example(self):
        # Create a new game using the Finkel rules.
        game = Game.create_finkel()

        while not game.is_finished():
            turn_player_name = game.get_turn().text_name

            if game.is_waiting_for_roll():
                # Roll the dice!
                roll = game.roll_dice()
                print(f"{turn_player_name}: Roll {roll.value}")

            else:
                # Make a random move.
                moves = game.find_available_moves()
                move = moves[random.randint(0, len(moves) - 1)]
                game.make_move(move)
                print(f"{turn_player_name}: {move.describe()}")

        # Report the winner!
        print(f"{game.get_winner().text_name} won the game!")



from enum import Enum
from typing import Optional


class PlayerType(Enum):
    """
    Represents the players of a game.
    """
    LIGHT = (1, "Light", 'L')
    """
    The light player.
    """

    DARK = (2, "Dark", 'D')
    """
    The dark player.
    """

    def __init__(self, value: int, text_name: str, character: str):
        self._value_ = value
        self._text_name = text_name
        self._character = character

    @property
    def text_name(self) -> str:
        """
        The name of this player.
        """
        return self._text_name

    @property
    def character(self) -> str:
        """
        The character used to represent this player in shorthand notations.
        """
        return self._character

    def get_other_player(self) -> 'PlayerType':
        """
        Retrieve the PlayerType representing the other player.
        """
        if self == PlayerType.LIGHT:
            return PlayerType.DARK
        elif self == PlayerType.DARK:
            return PlayerType.LIGHT
        else:
            raise ValueError(f"Unknown PlayerType {self}")

    @staticmethod
    def to_char(player: Optional['PlayerType']) -> str:
        """
        Convert the player to a single character used to represent the
        player in shorthand notations.
        """
        return player.character if player else '.'


class PlayerState:
    """
    A player state represents the state of a single player at a point in the game.
    This includes the player's score and number of pieces left to play.
    """
    __slots__ = ("_player", "_piece_count", "_score")

    _player: PlayerType
    _piece_count: int
    _score: int

    def __init__(self, player: PlayerType, piece_count: int, score: int):
        if piece_count < 0:
            raise ValueError("piece_count cannot be negative")
        if score < 0:
            raise ValueError("score cannot be negative")

        self._player = player
        self._piece_count = piece_count
        self._score = score

    @property
    def player(self) -> PlayerType:
        """
        The player that this state represents.
        """
        return self._player

    @property
    def piece_count(self) -> int:
        """
        The number of pieces that the player has available to introduce
        to the board.
        """
        return self._piece_count

    @property
    def score(self) -> int:
        """
        The number of pieces that the player has taken off the board.
        """
        return self._score

    def __hash__(self) -> int:
        return hash((self._player, self._piece_count, self._score))

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other):
            return False

        return self._player == other._player \
            and self._piece_count == other._piece_count and \
                self._score == other._score





class Tile:
    """
    Represents a position on or off the board.
    """
    __slots__ = ("_x", "_y", "_ix", "_iy")

    _x: int
    _y: int
    _ix: int
    _iy: int

    def __init__(self, x: int, y: int):
        if x < 1 or x > 26:
            raise ValueError(f"x must fall within the range [1, 26]. Invalid value: {x}")
        if y < 0:
            raise ValueError(f"y must not be negative. Invalid value: {y}");

        self._x = x
        self._y = y
        self._ix = x - 1
        self._iy = y - 1

    @property
    def x(self) -> int:
        """
        The x-coordinate of the tile. This coordinate is 1-based.
        """
        return self._x

    @property
    def y(self) -> int:
        """
        The y-coordinate of the tile. This coordinate is 1-based.
        """
        return self._y

    @property
    def ix(self) -> int:
        """
        The x-index of the tile. This coordinate is 0-based.
        """
        return self._ix

    @property
    def iy(self) -> int:
        """
        The y-index of the tile. This coordinate is 0-based.
        """
        return self._iy

    @staticmethod
    def from_indices(ix: int, iy: int) -> 'Tile':
        """
        Creates a new tile representing the tile at the
        indices (ix, iy), 0-based.
        """
        return Tile(ix + 1, iy + 1)

    def step_towards(self, other: 'Tile') -> 'Tile':
        """
        Takes a unit length step towards the other tile.
        """
        dx = other._x - self._x
        dy = other._y - self._y

        if abs(dx) + abs(dy) <= 1:
            return other

        if abs(dx) < abs(dy):
            return Tile(self._x, self._y + (1 if dy > 0 else -1))
        else:
            return Tile(self._x + (1 if dx > 0 else -1), self._y)

    def __hash__(self) -> int:
        return hash((self._x, self._y))

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False

        return self._x == other._x and self._y == other._y

    def __repr__(self) -> str:
        encoded_x = chr(self._x + (ord('A') - 1))
        return f"{encoded_x}{self._y}"

    def __str__(self) -> str:
        return repr(self)

    @staticmethod
    def from_string(encoded: str) -> 'Tile':
        """
        Decodes the tile coordinates from its encoded text (e.g., A4).
        """
        if len(encoded) < 2:
            raise ValueError("Incorrect format, expected at least two characters")

        x = ord(encoded[0]) - (ord('A') - 1)
        y = int(encoded[1:])
        return Tile(x, y)

    @staticmethod
    def create_list(*coordinates: list[tuple[int, int]]) -> list['Tile']:
        """
        Constructs a list of tiles from the tile coordinates.
        """
        return [Tile(x, y) for x, y in coordinates]

    @staticmethod
    def create_path(*coordinates: list[tuple[int, int]]) -> list['Tile']:
        """
        Constructs a path from waypoints on the board.
        """
        waypoints = Tile.create_list(*coordinates)
        if len(waypoints) == 0:
            raise ValueError("No coordinates provided")

        path = [waypoints[0]]
        for index in range(1, len(waypoints)):
            current = waypoints[index - 1]
            next = waypoints[index]
            while current != next:
                current = current.step_towards(next)
                path.append(current)

        return path



from .tile import Tile
from .path import BellPathPair, AsebPathPair
from enum import Enum
from typing import Iterable, Callable


class BoardShape:
    """
    Holds the shape of a board as a grid, and includes the location of all rosette tiles.
    """
    __slots__ = ("_name", "_tiles", "_rosettes", "_width", "_height")

    _name: str
    _tiles: set[Tile]
    _rosettes: set[Tile]
    _width: int
    _height: int

    def __init__(
            self,
            name: str,
            tiles: set[Tile],
            rosettes: set[Tile]
    ):

        if len(tiles) == 0:
            raise ValueError("A board shape requires at least one tile")

        for rosette in rosettes:
            if rosette not in tiles:
                raise ValueError(f"Rosette at {rosette} does not exist on the board")

        self._name = name
        self._tiles = tiles
        self._rosettes = rosettes

        x_values = [tile._x for tile in tiles]
        y_values = [tile._y for tile in tiles]

        min_x = min(x_values)
        min_y = min(y_values)
        if min_x != 1 or min_y != 1:
            raise ValueError(
                f"The board shape must be translated such that it has tiles "
                f"at an x-coordinate of 1, and at a y-coordinate of 1. "
                f"Minimum X = {min_x}, Minimum Y = {min_y}"
            )

        self._width = max(x_values)
        self._height = max(y_values)

    @property
    def name(self) -> str:
        """
        The name of this board shape.
        """
        return self._name

    @property
    def tiles(self) -> set[Tile]:
        """
        The set of tiles that fall within the bounds of this board shape.
        """
        return self._tiles

    @property
    def rosettes(self) -> set[Tile]:
        """
        The set of tiles that represent rosette tiles in this board shape.
        """
        return self._rosettes

    @property
    def width(self) -> int:
        """
        The number of x-coordinates that exist in this board shape.
        """
        return self._width

    @property
    def height(self) -> int:
        """
        The number of y-coordinates that exist in this board shape.
        """
        return self._height

    @property
    def area(self) -> int:
        """
        The number of tiles contained in this board shape.
        """
        return len(self._tiles)

    def contains(self, tile: Tile) -> bool:
        """
        Determines whether the given tile falls within this board shape.
        """
        return tile in self._tiles

    def contains_indices(self, ix: int, iy: int) -> bool:
        """
        Determines whether the tile at indices (ix, iy), 0-based,
        falls within the bounds of this shape of board.
        """
        if ix < 0 or iy < 0 or ix >= self._width or iy >= self._height:
            return False
        return self.contains(Tile.from_indices(ix, iy))

    def contains_all(self, tiles: Iterable[Tile]) -> bool:
        """
        Determines whether all provided tiles fall within this board shape.
        """
        for tile in tiles:
            if not self.contains(tile):
                return False
        return True

    def is_rosette(self, tile: Tile) -> bool:
        """
        Determines whether the given tile is a rosette tile in this board shape.
        """
        return tile in self._rosettes

    def is_rosette_indices(self, ix: int, iy: int) -> bool:
        """
        Determines whether the tile at the indices (ix, iy), 0-based,
        is a rosette tile in this board shape.
        """
        if ix < 0 or iy < 0 or ix >= self._width or iy >= self._height:
            return False
        return self.isRosette(Tile.from_indices(ix, iy))

    def is_equivalent(self, other: 'BoardShape') -> bool:
        """
        Determines whether this board shape covers the same tiles,
        and has the same rosettes, as other. This does not check
        that the names of the board shapes are the same.
        """
        return self._tiles == other._tiles and self._rosettes == other._rosettes

    def __eq__(self, other: 'BoardShape') -> bool:
        if type(other) is not type(self):
            return False
        return self.is_equivalent(other) and self._name == other._name


class AsebBoardShape(BoardShape):
    """
    The shape of the board used for the game Aseb.
    This board shape consists of 3 rows. The first row contains
    4 tiles, the second row contains 12 tiles, and the third
    also contains 4 tiles.
    """
    __slots__ = ()

    NAME: str = "Aseb"
    """
    The name given to this board shape.
    """

    BOARD_TILES: set[Tile] = set(AsebPathPair.LIGHT_PATH).union(set(AsebPathPair.DARK_PATH))
    """
    The set of all tiles that exist on the board.
    """

    ROSETTE_TILES: set[Tile] = {
        Tile(1, 1),
        Tile(3, 1),
        Tile(2, 4),
        Tile(2, 8),
        Tile(2, 12)
    }
    """
    The set of rosette tiles that exist on the board.
    """

    def __init__(self):
        super().__init__(
            AsebBoardShape.NAME,
            AsebBoardShape.BOARD_TILES,
            AsebBoardShape.ROSETTE_TILES
        )


class StandardBoardShape(BoardShape):
    """
    The standard shape of board used for The Royal Game of Ur that
    follows the game boards that were excavated by Sir Leonard Woolley.
    """
    __slots__ = ()

    NAME: str = "Standard"
    """
    The name given to this board shape.
    """

    BOARD_TILES: set[Tile] = set(BellPathPair.LIGHT_PATH).union(set(BellPathPair.DARK_PATH))
    """
    The set of all tiles that exist on the board.
    """

    ROSETTE_TILES: set[Tile] = {
        Tile(1, 1),
        Tile(3, 1),
        Tile(2, 4),
        Tile(1, 7),
        Tile(3, 7)
    }
    """
    The set of rosette tiles that exist on the board.
    """

    def __init__(self):
        super().__init__(
            StandardBoardShape.NAME,
            StandardBoardShape.BOARD_TILES,
            StandardBoardShape.ROSETTE_TILES
        )


class BoardType(Enum):
    """
    The type of board to use in a game.
    """

    STANDARD = (1, StandardBoardShape.NAME, lambda: StandardBoardShape())
    """
    The standard board shape.
    """

    ASEB = (2, AsebBoardShape.NAME, lambda: AsebBoardShape())
    """
    The Aseb board shape.
    """

    def __init__(self, value: int, text_name: str, create_board_shape: Callable[[], BoardShape]):
        self._value_ = value
        self._text_name = text_name
        self._create_board_shape = create_board_shape

    @property
    def text_name(self) -> str:
        """
        The name of this board shape.
        """
        return self._text_name

    def create_board_shape(self) -> BoardShape:
        """
        Create an instance of the board shape.
        """
        return self._create_board_shape()



from royalur.model import Board, PlayerType, PlayerState
from abc import ABC, abstractmethod
from overrides import overrides


class GameState(ABC):
    """
    A game state represents a single point within a game.
    """
    __slots__ = ("_board", "_light_player", "_dark_player")

    _board: Board
    _light_player: PlayerState
    _dark_player: PlayerState

    def __init__(
            self,
            board: Board,
            light_player: PlayerState,
            dark_player: PlayerState
    ):
        super().__init__()
        self._board = board
        self._light_player = light_player
        self._dark_player = dark_player

    @property
    def board(self) -> Board:
        """
        The state of the pieces on the board.
        """
        return self._board

    @property
    def light_player(self) -> PlayerState:
        """
        The state of the light player.
        """
        return self._light_player

    @property
    def dark_player(self) -> PlayerState:
        """
        The state of the dark player.
        """
        return self._dark_player

    def get_player(self, player: PlayerType) -> PlayerState:
        """
        Gets the state of the given player.
        """
        if player == PlayerType.LIGHT:
            return self._light_player
        if player == PlayerType.DARK:
            return self._dark_player

        raise ValueError(f"Unknown player {player}")

    @abstractmethod
    def is_playable(self) -> bool:
        """
        Returns whether this state is a valid state to be played from.
        """
        pass

    @abstractmethod
    def is_finished(self) -> bool:
        """
        Returns whether this state represents a finished game.
        """
        pass

    @abstractmethod
    def describe(self) -> str:
        """
        Generates an English text description of the state of the game.
        """
        pass


class OngoingGameState(GameState):
    """
    A game state from within a game where a winner has not yet been determined.
    """
    __slots__ = ("_turn",)

    _turn: PlayerType

    def __init__(
            self,
            board: Board,
            light_player: PlayerState,
            dark_player: PlayerState,
            turn: PlayerType
    ):
        super().__init__(board, light_player, dark_player)
        self._turn = turn

    @overrides
    def is_finished(self) -> bool:
        return False

    def get_turn(self) -> PlayerType:
        """
        Gets the player who made an action or that should make an action.
        """
        return self._turn

    def get_waiting(self) -> PlayerType:
        """
        Gets the player that is waiting whilst the other player makes the
        next interaction with the game.
        """
        return self._turn.get_other_player()

    def get_turn_player(self) -> PlayerState:
        """
        Gets the state of the player that we are waiting on to interact with the game.
        """
        return self.get_player(self._turn)

    def get_waiting_player(self) -> PlayerState:
        """
        Gets the state of the player that is waiting whilst the other player makes the
        next interaction with the game.
        """
        return self.get_player(self.get_waiting())


class WinGameState(GameState):
    """
    A game state where a player has won the game.
    """
    __slots__ = ("_winner",)

    _winner: PlayerType

    def __init__(
            self,
            board: Board,
            light_player: PlayerState,
            dark_player: PlayerState,
            winner: PlayerType
    ):
        super().__init__(board, light_player, dark_player)
        self._winner = winner

    @overrides
    def is_playable(self) -> bool:
        return False

    @overrides
    def is_finished(self) -> bool:
        return True

    def get_winner(self) -> PlayerType:
        """
        Gets the player that won the game.
        """
        return self._winner

    def get_loser(self) -> PlayerType:
        """
        Gets the player that lost the game.
        """
        return self._winner.get_other_player()

    def get_winning_player(self) -> PlayerState:
        """
        Gets the state of the player that won the game.
        """
        return self.get_player(self._winner)

    def get_losing_player(self) -> PlayerState:
        """
        Gets the state of the player that lost the game.
        """
        return self.get_player(self.get_loser())

    @overrides
    def describe(self) -> str:
        winner = self._winner.text_name.lower()
        return f"The {winner} player has won!"



"""
This package contains the model for storing the state
of games of the Royal Game of Ur.
"""

from .tile import Tile
from .player import PlayerType, PlayerState
from .path import (
    PathPair, PathType,
    AsebPathPair, BellPathPair,
    MastersPathPair, MurrayPathPair,
    SkiriukPathPair,
)
from .shape import (
    BoardShape, BoardType,
    StandardBoardShape, AsebBoardShape,
)
from .board import Piece, Move, Board
from .dice import (
    Roll, Dice, DiceType,
    BinaryDice, BinaryDice0AsMax,
)
from .settings import GameSettings
from .metadata import GameMetadata



from royalur.model import Board, PlayerState, PlayerType, Roll, Move
from .state import OngoingGameState
from overrides import overrides


class PlayableGameState(OngoingGameState):
    """
    A game state where we are waiting for interactions from a player.
    """
    __slots__ = ()

    @overrides
    def is_playable(self) -> bool:
        return True


class WaitingForRollGameState(PlayableGameState):
    """
    A game state where the game is waiting for a player to roll the dice.
    """

    @overrides
    def describe(self) -> str:
        turn = self.get_turn().text_name.lower()
        return f"Waiting for the {turn} player to roll the dice."


class WaitingForMoveGameState(PlayableGameState):
    """
    A game state where the game is waiting for a player to roll the dice.
    """
    __slots__ = ("_roll", "_available_moves")

    _roll: Roll
    _available_moves: list[Move]

    def __init__(
            self,
            board: Board,
            light_player: PlayerState,
            dark_player: PlayerState,
            turn: PlayerType,
            roll: Roll,
            available_moves: list[Move]
    ):
        super().__init__(board, light_player, dark_player, turn)
        self._roll = roll
        self._available_moves = available_moves

    @property
    def roll(self) -> Roll:
        """
        The roll that represents the number of places the player can move a piece.
        """
        return self._roll

    @property
    def available_moves(self) -> list[Move]:
        """
        The moves that are available to be made from this position.
        """
        return self._available_moves

    @overrides
    def describe(self) -> str:
        turn = self.get_turn().text_name.lower()
        return f"Waiting for the {turn} player to make a " \
            f"move with their roll of {self._roll.value}"



"""
This package contains representations of the states of games.
"""

from .state import (
    GameState,
    OngoingGameState,
    WinGameState,
)
from .action import (
    ActionGameState,
    RolledGameState,
    MovedGameState,
)
from .playable import (
    PlayableGameState,
    WaitingForRollGameState,
    WaitingForMoveGameState,
)



from .shape import BoardShape, BoardType
from .path import PathPair, PathType
from .dice import Dice, DiceType
from typing import Union, Callable


class GameSettings:
    """
    Settings for running games of the Royal Game of Ur. This is built for
    convenience, and cannot represent all possible combinations of rules
    that can be used to play the Royal Game of Ur. If a more exotic set of
    rules is desired, then you will need to construct your games manually.
    """
    __slots__ = (
        "_board_shape", "_paths", "_dice_factory", "_starting_piece_count", "_safe_rosettes",
        "_rosettes_grant_extra_rolls", "_captures_grant_extra_rolls"
    )

    _board_shape: BoardShape
    _paths: PathPair
    _dice_factory: Callable[[], Dice]
    _starting_piece_count: int
    _safe_rosettes: bool
    _rosettes_grant_extra_rolls: bool
    _captures_grant_extra_rolls: bool

    def __init__(
            self,
            board_shape: BoardShape,
            paths: PathPair,
            dice_factory: Callable[[], Dice],
            starting_piece_count: int,
            safe_rosettes: bool,
            rosettes_grant_extra_rolls: bool,
            captures_grant_extra_rolls: bool,
    ):
        if starting_piece_count < 1:
            raise ValueError("starting piece count must be at least 1")

        self._board_shape = board_shape
        self._paths = paths
        self._dice_factory = dice_factory
        self._starting_piece_count = starting_piece_count
        self._safe_rosettes = safe_rosettes
        self._rosettes_grant_extra_rolls = rosettes_grant_extra_rolls
        self._captures_grant_extra_rolls = captures_grant_extra_rolls

    @staticmethod
    def create_finkel() -> 'GameSettings':
        """
        Creates an instance of the rules used in the YouTube
        video Tom Scott vs. Irving Finkel.
        """
        return GameSettings(
            BoardType.STANDARD.create_board_shape(),
            PathType.BELL.create_path_pair(),
            lambda: DiceType.FOUR_BINARY.create_dice(),
            7,
            True,
            True,
            False,
        )

    @staticmethod
    def create_masters() -> 'GameSettings':
        """
        Creates an instance of the settings proposed by James Masters.
        """
        return GameSettings(
            BoardType.STANDARD.create_board_shape(),
            PathType.MASTERS.create_path_pair(),
            lambda: DiceType.FOUR_BINARY.create_dice(),
            7,
            False,
            True,
            False,
        )

    @staticmethod
    def create_aseb() -> 'GameSettings':
        """
        Creates an instance of the settings used for Aseb.
        """
        return GameSettings(
            BoardType.ASEB.create_board_shape(),
            PathType.ASEB.create_path_pair(),
            lambda: DiceType.FOUR_BINARY.create_dice(),
            5,
            True,
            True,
            False,
        )

    @property
    def board_shape(self) -> BoardShape:
        """
        The shape of the game board.
        """
        return self._board_shape

    def with_board_shape(self, board_shape: Union[BoardShape, BoardType]) -> 'GameSettings':
        """
        Generates new game settings with a new board shape.
        """
        if isinstance(board_shape, BoardType):
            board_shape = board_shape.create_board_shape()

        return GameSettings(
            board_shape, self._paths, self._dice_factory,
            self._starting_piece_count, self._safe_rosettes,
            self._rosettes_grant_extra_rolls,
            self._captures_grant_extra_rolls
        )

    @property
    def paths(self) -> PathPair:
        """
        The paths that each player must take around the board.
        """
        return self._paths

    def with_paths(self, paths: Union[PathPair, PathType]) -> 'GameSettings':
        """
        Generates new game settings with new paths.
        """
        if isinstance(paths, PathType):
            paths = paths.create_path_pair()

        return GameSettings(
            self._board_shape, paths, self._dice_factory,
            self._starting_piece_count, self._safe_rosettes,
            self._rosettes_grant_extra_rolls,
            self._captures_grant_extra_rolls
        )

    @property
    def dice(self) -> Callable[[], Dice]:
        """
        A generator for the dice that should be used to generate dice rolls in games.
        """
        return self._dice_factory

    def with_dice(self, dice: Union[Callable[[], Dice], DiceType]) -> 'GameSettings':
        """
        Generates new game settings with new dice.
        """
        if isinstance(dice, DiceType):
            dice = lambda: dice.create_dice()

        return GameSettings(
            self._board_shape, self._paths, dice,
            self._starting_piece_count, self._safe_rosettes,
            self._rosettes_grant_extra_rolls,
            self._captures_grant_extra_rolls
        )

    @property
    def starting_piece_count(self) -> int:
        """
        The number of pieces that each player starts with.
        """
        return self._starting_piece_count

    def with_starting_piece_count(self, starting_piece_count: int) -> 'GameSettings':
        """
        Generates new game settings with a new starting piece count.
        """
        return GameSettings(
            self._board_shape, self._paths, self._dice_factory,
            starting_piece_count, self._safe_rosettes,
            self._rosettes_grant_extra_rolls,
            self._captures_grant_extra_rolls
        )

    @property
    def safe_rosettes(self) -> bool:
        """
        Whether pieces on rosette tiles are safe from capture.
        """
        return self._safe_rosettes

    def with_safe_rosettes(self, safe_rosettes: bool) -> 'GameSettings':
        """
        Generates new game settings with a new value for whether rosettes are safe.
        """
        return GameSettings(
            self._board_shape, self._paths, self._dice_factory,
            self._starting_piece_count, safe_rosettes,
            self._rosettes_grant_extra_rolls,
            self._captures_grant_extra_rolls
        )

    @property
    def rosettes_grant_extra_rolls(self) -> bool:
        """
        Whether landing on a rosette tile grants the player an
        additional roll of the dice.
        """
        return self._rosettes_grant_extra_rolls

    def with_rosettes_grant_extra_rolls(
            self, rosettes_grant_extra_rolls: bool
    ) -> 'GameSettings':
        """
        Generates new game settings with a new value for whether
        landing on a rosette grants an extra roll.
        """
        return GameSettings(
            self._board_shape, self._paths, self._dice_factory,
            self._starting_piece_count, self._safe_rosettes,
            rosettes_grant_extra_rolls,
            self._captures_grant_extra_rolls
        )

    @property
    def captures_grant_extra_rolls(self) -> bool:
        """
        Whether capturing a piece grants the player an additional
        roll of the dice.
        """
        return self._captures_grant_extra_rolls

    def with_captures_grant_extra_rolls(
            self, captures_grant_extra_rolls: bool
    ) -> 'GameSettings':
        """
        Generates new game settings with a new value for whether
        capturing a piece grants an extra roll.
        """
        return GameSettings(
            self._board_shape, self._paths, self._dice_factory,
            self._starting_piece_count, self._safe_rosettes,
            self._rosettes_grant_extra_rolls,
            captures_grant_extra_rolls
        )



from royalur.model import Board, PlayerState, PlayerType, Roll, Move
from .state import OngoingGameState
from overrides import overrides


class ActionGameState(OngoingGameState):
    """
    A game state that is included in the middle of a game to record an action that
    was taken, but that is not a valid state to be in.
    """
    __slots__ = ()

    @overrides
    def is_playable(self) -> bool:
        return False


class RolledGameState(ActionGameState):
    """
    A game state that represents a roll that was made in a game.
    """
    __slots__ = ("_roll", "_available_moves")

    _roll: Roll
    _available_moves: list[Move]

    def __init__(
            self,
            board: Board,
            light_player: PlayerState,
            dark_player: PlayerState,
            turn: PlayerType,
            roll: Roll,
            available_moves: list[Move]
    ):
        super().__init__(board, light_player, dark_player, turn)
        self._roll = roll
        self._available_moves = available_moves

    @property
    def roll(self) -> Roll:
        """
        The roll of the dice that was used for the move.
        """
        return self._roll

    @property
    def available_moves(self) -> list[Move]:
        """
        The moves that are available from this position using the given roll.
        """
        return self._available_moves

    @overrides
    def describe(self) -> str:
        builder = []
        builder.append("The ")
        builder.append(self.get_turn().text_name.lower())
        builder.append(" player rolled ")
        builder.append(str(self.roll))

        if not self._available_moves:
            if self._roll.value() == 0:
                builder.append(", and had no moves")
            else:
                builder.append(", and all moves were blocked")

        builder.append(".")
        return "".join(builder)


class MovedGameState(ActionGameState):
    """
    A game state that represents a move of a piece on the board.
    """
    __slots__ = ("_roll", "_move")

    _roll: Roll
    _move: Move

    def __init__(
            self,
            board: Board,
            light_player: PlayerState,
            dark_player: PlayerState,
            turn: PlayerType,
            roll: Roll,
            move: Move
    ):
        super().__init__(board, light_player, dark_player, turn)
        self._roll = roll
        self._move = move

    @property
    def roll(self) -> Roll:
        """
        The roll of the dice that was used for the move.
        """
        return self._roll

    @property
    def move(self) -> Move:
        """
        The move that was made.
        """
        return self._move

    @overrides
    def describe(self) -> str:
        builder = []
        builder.append("The ")
        builder.append(self.get_turn().text_name.lower())
        builder.append(" player ")

        introducing = self.move.is_introducing_piece()
        scoring = self.move.is_scoring_piece()

        if introducing and scoring:
            builder.append("scored a newly introduced piece.")
        elif scoring:
            builder.append("scored their ")
            builder.append(str(self.move.get_source()))
            builder.append("piece.")
        elif introducing:
            builder.append("introduced a piece to ")
            builder.append(str(self.move.get_dest()))
            builder.append(".")
        else:
            builder.append("moved their ")
            builder.append(str(self.move.get_source()))
            builder.append(" piece to ")
            if self.move.is_capture():
                builder.append("capture ")
            builder.append(str(self.move.get_dest()))
            builder.append(".")

        return "".join(builder)



from royalur.model import (
    Board, BoardShape, GameMetadata, GameSettings,
    Move, PathPair, Piece, Dice,
    PlayerType, PlayerState, Roll
)
from royalur.rules.state import (
    GameState,
    WaitingForRollGameState,
    WaitingForMoveGameState,
    RolledGameState,
    MovedGameState,
    WinGameState,
)
from royalur.rules.rules import (
    PieceProvider, PlayerStateProvider,
    RuleSet, RuleSetProvider,
)
from overrides import overrides
from typing import Callable


class SimplePieceProvider(PieceProvider):
    """
    Provides new instances of, and manipulations to, simple pieces.
    """
    __slots__ = ()

    @overrides
    def create_introduced(self, owner: PlayerType, new_path_index: int) -> Piece:
        return Piece(owner, new_path_index)

    @overrides
    def create_moved(self, origin_piece: Piece, new_path_index: int) -> Piece:
        return Piece(origin_piece.owner, new_path_index)


class SimplePlayerStateProvider(PlayerStateProvider):
    """
    Provides new instances of, and manipulations to, simple player states.
    """
    __slots__ = ("_starting_piece_count",)

    _starting_piece_count: int

    def __init__(self, starting_piece_count: int):
        super().__init__()
        self._starting_piece_count = starting_piece_count

    @overrides
    def get_starting_piece_count(self) -> int:
        return self._starting_piece_count

    @overrides
    def create(self, player: PlayerType):
        return PlayerState(player, self._starting_piece_count, 0)

    @overrides
    def apply_piece_introduced(self, player_state: PlayerState, piece: Piece):
        return PlayerState(
            player_state.player,
            player_state.piece_count - 1,
            player_state.score
        )

    @overrides
    def apply_piece_captured(self, player_state: PlayerState, piece: Piece):
        return PlayerState(
            player_state.player,
            player_state.piece_count + 1,
            player_state.score
        )

    @overrides
    def apply_piece_scored(self, player_state: PlayerState, piece: Piece):
        return PlayerState(
            player_state.player,
            player_state.piece_count,
            player_state.score + 1
        )


class SimpleRuleSet(RuleSet):
    """
    The most common, simple, rules of the Royal Game of Ur.
    This still allows a large range of custom rules.
    """
    __slots__ = ("_safe_rosettes", "_rosettes_grant_extra_rolls", "_captures_grant_extra_rolls")

    _safe_rosettes: bool
    _rosettes_grant_extra_rolls: bool
    _captures_grant_extra_rolls: bool

    def __init__(
            self,
            board_shape: BoardShape,
            paths: PathPair,
            dice_factory: Callable[[], Dice],
            piece_provider: PieceProvider,
            player_state_provider: PlayerStateProvider,
            safe_rosettes: bool,
            rosettes_grant_extra_rolls: bool,
            captures_grant_extra_rolls: bool,
    ):
        super().__init__(board_shape, paths, dice_factory, piece_provider, player_state_provider)
        self._safe_rosettes = safe_rosettes
        self._rosettes_grant_extra_rolls = rosettes_grant_extra_rolls
        self._captures_grant_extra_rolls = captures_grant_extra_rolls

    @overrides
    def are_rosettes_safe(self) -> bool:
        return self._safe_rosettes

    @overrides
    def do_rosettes_grant_extra_rolls(self) -> bool:
        return self._rosettes_grant_extra_rolls

    @overrides
    def do_captures_grant_extra_rolls(self) -> bool:
        return self._captures_grant_extra_rolls

    @overrides
    def generate_initial_game_state(self) -> GameState:
        return WaitingForRollGameState(
            Board(self._board_shape),
            self._player_state_provider.create(PlayerType.LIGHT),
            self._player_state_provider.create(PlayerType.DARK),
            PlayerType.LIGHT,
        )

    @overrides
    def find_available_moves(
        self,
        board: Board,
        player: PlayerState,
        roll: Roll
    ) -> list[Move]:

        if roll.value == 0:
            return []

        player_type = player.player
        path = self._paths.get(player_type)
        moves = []

        # Check if a piece can be taken off the board.
        if roll.value <= len(path):
            score_path_index = len(path) - roll.value
            score_tile = path[score_path_index]
            score_piece = board.get(score_tile)
            if score_piece is not None \
                and score_piece.owner == player_type \
                and score_piece.path_index == score_path_index:

                moves.append(Move(
                    player_type, score_tile, score_piece,
                    None, None, None
                ))

        # Check for pieces on the board that can be moved to another tile on the board.
        for path_index in range(-1, len(path) - roll.value):

            if path_index >= 0:
                # Move a piece on the board.
                tile = path[path_index]
                piece = board.get(tile)
                if piece is None \
                    or piece.owner != player_type \
                    or piece.path_index != path_index:
                    continue

            elif player.piece_count > 0:
                # Introduce a piece to the board.
                tile = None
                piece = None

            else:
                continue

            # Check if the destination is free.
            dest_path_index = path_index + roll.value
            dest = path[dest_path_index]
            dest_piece = board.get(dest)

            if dest_piece is not None:
                # Cannot capture your own piece.
                if dest_piece.owner == player_type:
                    continue

                # Can't capture pieces on rosettes if they are safe.
                if self._safe_rosettes and self._board_shape.is_rosette(dest):
                    continue

            # Generate the move.
            if path_index >= 0:
                moved_piece = self._piece_provider.create_moved(piece, dest_path_index)
            else:
                moved_piece = self._piece_provider.create_introduced(player_type, dest_path_index)

            moves.append(Move(
                player_type, tile, piece, dest, moved_piece, dest_piece
            ))

        return moves

    @overrides
    def apply_roll(
            self,
            state: WaitingForRollGameState,
            roll: Roll
    ) -> list[GameState]:

        available_moves = self.find_available_moves(
            state.board, state.get_turn_player(), roll
        )
        rolled_state = RolledGameState(
            state.board, state.light_player, state.dark_player,
            state.get_turn(), roll, available_moves
        )

        # Swap turn when rolling the player has no available moves.
        if len(available_moves) == 0:
            new_turn = state.get_turn().get_other_player()
            return [rolled_state, WaitingForRollGameState(
                state.board, state.light_player, state.dark_player, new_turn
            )]

        # The player has available moves.
        return [rolled_state, WaitingForMoveGameState(
            state.board, state.light_player, state.dark_player,
            state.get_turn(), roll, available_moves
        )]

    def should_grant_extra_roll(self, moved_state: MovedGameState) -> bool:
        """
        Determines whether the move represent by the given moved state
        should grant another roll to the player that made the move.
        """
        move = moved_state.move

        if self._rosettes_grant_extra_rolls and move.is_dest_rosette(self._board_shape):
            return True

        return self._captures_grant_extra_rolls and move.is_capture()

    @overrides
    def apply_move(
            self,
            state: WaitingForMoveGameState,
            move: Move
    ) -> list[GameState]:

        # Generate the state representing the move that was made.
        moved_state = MovedGameState(
            state.board, state.light_player, state.dark_player,
            state.get_turn(), state.roll, move
        )

        # Apply the move to the board.
        board = state.board.copy()
        move.apply(board)

        # Apply the move to the player that made the move.
        turn_player = state.get_turn_player()

        if move.is_introducing_piece():
            introduced_piece = move.get_dest_piece()
            turn_player = self._player_state_provider.apply_piece_introduced(
                turn_player, introduced_piece
            )

        elif move.is_scoring_piece():
            scored_piece = move.get_source_piece()
            turn_player = self._player_state_provider.apply_piece_scored(
                turn_player, scored_piece
            )

        # Apply the effects of the move to the other player.
        other_player = state.get_waiting_player()
        if move.is_capture():
            captured_piece = move.get_captured_piece()
            other_player = self._player_state_provider.apply_piece_captured(
                other_player, captured_piece
            )

        # Determine which player is which.
        turn = turn_player.player
        light_player = (turn_player if turn == PlayerType.LIGHT else other_player)
        dark_player = (turn_player if turn == PlayerType.DARK else other_player)

        # Check if the player has won the game.
        turn_player_pieces = turn_player.piece_count
        if move.is_scoring_piece() and turn_player_pieces + board.count_pieces(turn) <= 0:
            return [moved_state, WinGameState(
                board, light_player, dark_player, turn
            )]

        # Determine whose turn it will be in the next state.
        grant_extra_roll = self.should_grant_extra_roll(moved_state)
        next_turn = (turn if grant_extra_roll else turn.get_other_player())
        return [moved_state, WaitingForRollGameState(
            board, light_player, dark_player, next_turn
        )]


class SimpleRuleSetProvider(RuleSetProvider):
    """
    A provider that creates simple rule sets.
    """
    __slots__ = ()

    @overrides
    def create(
            self,
            settings: GameSettings,
            metadata: GameMetadata
    ) -> RuleSet:

        piece_provider = SimplePieceProvider()
        player_state_provider = SimplePlayerStateProvider(
            settings.starting_piece_count
        )
        return SimpleRuleSet(
            settings.board_shape,
            settings.paths,
            settings.dice,
            piece_provider,
            player_state_provider,
            settings.safe_rosettes,
            settings.rosettes_grant_extra_rolls,
            settings.captures_grant_extra_rolls
        )



"""
This package contains the rules to simulate games
of the Royal Game of Ur.
"""

from .state import (
    GameState,
    ActionGameState,
    PlayableGameState,
    WaitingForRollGameState,
    WaitingForMoveGameState,
    RolledGameState,
    MovedGameState,
    WinGameState
)
from .rules import (
    PieceProvider, PlayerStateProvider,
    RuleSet, RuleSetProvider,
)
from .simple import (
    SimplePieceProvider, SimplePlayerStateProvider,
    SimpleRuleSet, SimpleRuleSetProvider,
)



from .model import (
    GameMetadata,
    GameSettings,
    DiceType,
    Dice, Roll,
    BoardShape,
    BoardType,
    PathPair,
    PathType,
    Board, Move,
    Piece, Tile,
    PlayerState,
    PlayerType,
)
from .rules import (
    RuleSet,
    RuleSetProvider,
    GameState,
    ActionGameState,
    PlayableGameState,
    WaitingForRollGameState,
    WaitingForMoveGameState,
    MovedGameState,
    WinGameState,
)
from .rules.simple import SimpleRuleSetProvider
from typing import Iterable, Callable


class Game:
    """
    A game of the Royal Game of Ur. Provides methods to allow the playing of games,
    and methods to support the retrieval of history about the moves that were made.
    """
    __slots__ = ("_rules", "_dice", "_metadata", "_states")

    _rules: RuleSet
    _dice: Dice
    _metadata: GameMetadata
    _states: list[GameState]

    def __init__(
            self,
            rules: RuleSet,
            metadata: GameMetadata | None = None,
            states: list[GameState] | None = None
    ):
        if metadata is None:
            metadata = GameMetadata.create_for_new_game(rules.settings)
        if states is None:
            states = [rules.generate_initial_game_state()]

        if len(states) == 0:
            raise ValueError("Games must have at least one state")

        self._rules = rules
        self._metadata = metadata
        self._dice = rules.dice_factory()
        self._states = []

        self.add_states(states)

    def copy(self) -> 'Game':
        """
        Create an exact copy of this game.
        """
        if type(self) != Game:
            raise RuntimeError(f"{type(self)} does not support copy")

        new_game = Game(
            self._rules, self._metadata.copy(), self._states
        )
        new_game._dice.copy_from(self._dice)
        return new_game

    def add_states(self, states: Iterable[GameState]):
        """
        Adds all the given states to this game.
        """
        seen = 0
        for state in states:
            seen += 1
            self.add_state(state)

        if seen == 0:
            raise ValueError("There were no states to add")

    def add_state(self, state: GameState):
        """
        Adds the given state to this game.
        """
        self._states.append(state)

    @property
    def rules(self) -> RuleSet:
        """
        The set of rules that are being used for this game.
        """
        return self._rules

    @property
    def metadata(self) -> GameMetadata:
        """
        The metadata of this game.
        """
        return self._metadata

    @property
    def dice(self) -> Dice:
        """
        The dice to be used to make dice rolls.
        """
        return self._dice

    @property
    def states(self) -> list[GameState]:
        """
        The states that have occurred so far in the game.
        """
        return [*self._states]

    def get_current_state(self) -> GameState:
        """
        Gets the current state of the game.
        """
        return self._states[-1]

    def get_action_states(self) -> list[ActionGameState]:
        """
        Gets the states that represent the actions that have been
        made so far in the game. The last state in the list represents
        the last action that was taken in this game.
        """
        return [state for state in self._states if isinstance(state, ActionGameState)]

    def get_landmark_states(self) -> list[ActionGameState]:
        """
        Gets all moves that were made in the game, as well as
        the current state of the game. These states are considered
        landmark states as they contain all the information required
        to recreate everything that happened in the game so far.
        """
        return [
            state for index, state in enumerate(self._states)
            if isinstance(state, MovedGameState) or index == len(self._states) - 1
        ]

    def is_playable(self) -> bool:
        """
        Determines whether the game is currently in a playable state.
        """
        return isinstance(self.get_current_state(), PlayableGameState)

    def is_waiting_for_roll(self) -> bool:
        """
        Determines whether the game is currently waiting for a roll from a player.
        """
        return isinstance(self.get_current_state(), WaitingForRollGameState)

    def is_waiting_for_move(self) -> bool:
        """
        Determines whether the game is currently waiting for a move from a player.
        """
        return isinstance(self.get_current_state(), WaitingForMoveGameState)

    def is_finished(self) -> bool:
        """
        Determines whether the game is currently in a finished state.
        """
        return isinstance(self.get_current_state(), WinGameState)

    def get_current_playable_state(self) -> PlayableGameState:
        """
        Gets the current state of this game as a PlayableGameState.
        This will throw an error if the game is not in a playable state.
        """
        state = self.get_current_state()
        if not isinstance(state, PlayableGameState):
            raise RuntimeError("This game is not in a playable state")

        return state

    def get_current_waiting_for_roll_state(self) -> WaitingForRollGameState:
        """
        Gets the current state of this game as an instance of WaitingForRollGameState.
        This will throw an error if the game is not waiting for a roll from a player.
        """
        state = self.get_current_state()
        if not isinstance(state, WaitingForRollGameState):
            raise RuntimeError("This game is not in a waiting for roll state")

        return state

    def get_current_waiting_for_move_state(self) -> WaitingForMoveGameState:
        """
        Gets the current state of this game as an instance of WaitingForMoveGameState.
        This will throw an error if the game is not waiting for a  move from a player.
        """
        state = self.get_current_state()
        if not isinstance(state, WaitingForMoveGameState):
            raise RuntimeError("This game is not in a waiting for move state")

        return state

    def get_current_win_state(self) -> WinGameState:
        """
        Gets the current state of this game as an instance of WinGameState.
        This will throw an error if the game has not been won.
        """
        state = self.get_current_state()
        if not isinstance(state, WinGameState):
            raise RuntimeError("This game is not in a win state")

        return state

    def roll_dice(self, value: Roll | int | None = None) -> Roll:
        """
        Rolls the dice, and updates the state of the game accordingly.
        If a value is supplied, then the roll will have that value.
        """
        if value is None or not isinstance(value, Roll):
            roll = self._dice.roll(value)
        else:
            roll = value

        # Update the state of the game after rolling.
        state = self.get_current_waiting_for_roll_state()
        self.add_states(self._rules.apply_roll(state, roll))
        return roll

    def find_available_moves(self) -> list[Move]:
        """
        Finds all moves that can be made from the current position.
        """
        return self.get_current_waiting_for_move_state().available_moves

    def _make_move(self, move: Move):
        """
        Applies a move.
        """
        state = self.get_current_waiting_for_move_state()
        self.add_states(self._rules.apply_move(state, move))


    def make_move(self, move: Move | Piece | Tile):
        """
        Applies the given move to update the state of the game.
        If a piece or tile is provided, this will look for a move that
        moves the given piece, or a piece from the given tile.
        If a move is provided, this does not check whether it is valid.
        """
        if isinstance(move, Move):
            self._make_move(move)
            return

        # Find the move for the given piece.
        if isinstance(move, Piece):
            piece = move
            state = self.get_current_waiting_for_move_state()
            for avail_move in state.available_moves:
                if avail_move.has_source() and avail_move.get_source_piece() == piece:
                    self._make_move(avail_move)
                    return

            raise ValueError(f"The piece cannot be moved, {piece}")

        # Find the move for the given tile.
        if isinstance(move, Tile):
            tile = move
            paths = self._rules.paths
            state = self.get_current_waiting_for_move_state()
            for avail_move in state.available_moves:
                if avail_move.get_source(paths) == tile:
                    self._make_move(avail_move)
                    return

            raise ValueError(f"There is no available move from {piece}")

        raise RuntimeError("move is not a Move, Piece, or Tile")

    def make_move_introducing_piece(self):
        """
        Moves a new piece onto the board.
        """
        state = self.get_current_waiting_for_move_state()
        for move in state.available_moves:
            if move.is_introducing_piece():
                self._make_move(move)
                return

        raise ValueError(f"There is no available move to introduce a piece to the board")

    def get_board(self) -> Board:
        """
        Gets the current state of the board.
        """
        return self.get_current_state().board

    def get_light_player(self) -> PlayerState:
        """
        Gets the current state of the light player.
        """
        return self.get_current_state().light_player

    def get_dark_player(self) -> PlayerState:
        """
        Gets the current state of the dark player.
        """
        return self.get_current_state().dark_player

    def get_player(self, player: PlayerType) -> PlayerState:
        """
        Gets the current state of the given player.
        """
        return self.get_current_state().get_player(player)

    def get_turn(self) -> PlayerType:
        """
        Gets the player who can make the next interaction with the game.
        """
        return self.get_current_playable_state().get_turn()

    def get_turn_or_winner(self) -> PlayerType:
        """
        Gets the player who can make the next interaction with the game,
        or the winner of the game if it is finished.
        """
        state = self.get_current_state()
        if isinstance(state, PlayableGameState):
            return state.get_turn()

        if isinstance(state, WinGameState):
            return state.get_winner()

        raise RuntimeError("The game is not in a playable or won state")

    def get_turn_player(self) -> PlayerState:
        """
        Gets the state of the player whose turn it is.
        """
        return self.get_current_playable_state().get_turn_player()

    def get_waiting_player(self) -> PlayerState:
        """
        Gets the state of the player that is waiting as it is not their turn.
        """
        return self.get_current_playable_state().get_waiting_player()

    def get_winner(self) -> PlayerType:
        """
        Gets the player that won the game.
        """
        return self.get_current_win_state().get_winner()

    def get_loser(self) -> PlayerType:
        """
        Gets the player that lost the game.
        """
        return self.get_current_win_state().get_loser()

    def get_winning_player(self) -> PlayerState:
        """
        Gets the state of the winning player.
        """
        return self.get_current_win_state().get_winning_player()

    def get_losing_player(self) -> PlayerState:
        """
        Gets the state of the losing player.
        """
        return self.get_current_win_state().get_losing_player()

    def get_roll(self) -> Roll:
        """
        Gets the roll that was made that can be used by the
        current turn player to make a move.
        """
        return self.get_current_waiting_for_move_state().roll

    @staticmethod
    def builder() -> 'GameBuilder':
        """
        Creates a builder to assist in constructing games with custom settings.
        """
        return GameBuilder(GameSettings.create_finkel(), SimpleRuleSetProvider())

    @staticmethod
    def create(settings: GameSettings):
        """
        Creates a game that follows the given settings.
        """
        return Game.builder().replace_settings(settings).build()

    @staticmethod
    def create_finkel():
        """
        Creates a game that follows the rules proposed by Irving Finkel.
        This uses the simple rules, the standard board shape, Bell's path, safe
        rosette tiles, the standard dice, and seven starting pieces per player.
        """
        return Game.create(GameSettings.create_finkel())

    @staticmethod
    def create_finkel():
        """
        Creates a game that follows the rules proposed by James Masters.
        This uses the simple rules, the standard board shape, Bell's path, unsafe
        rosette tiles, the standard dice, and seven starting pieces per player.
        """
        return Game.create(GameSettings.create_masters())

    @staticmethod
    def create_aseb():
        """
        Creates a game of Aseb. This uses the simple rules, the Aseb board shape,
        the Aseb paths, the standard dice, and five starting pieces per player.
        """
        return Game.create(GameSettings.create_aseb())


class GameBuilder:
    """
    A builder to help in the creation of custom games of the Royal Game of Ur.
    """
    __slots__ = ("_settings", "_rule_set_provider")

    _settings: GameSettings
    _rule_set_provider: RuleSetProvider

    def __init__(
            self,
            settings: GameSettings,
            rule_set_provider: RuleSetProvider
    ):
        self._settings = settings
        self._rule_set_provider = rule_set_provider

    @property
    def settings(self) -> GameSettings:
        """
        The settings of the game being built.
        """
        return self._settings

    @property
    def rule_set_provider(self) -> RuleSetProvider:
        """
        The provider to use to construct the final rule set.
        """
        return self._rule_set_provider

    def replace_settings(self, settings: GameSettings) -> 'GameBuilder':
        """
        Create a copy of this game builder with new settings.
        """
        return GameBuilder(settings, self.rule_set_provider)

    def finkel(self) -> 'GameBuilder':
        """
        Creates a new builder that allows the construction of games
        following the simple rules proposed by Irving Finkel.
        """
        return self.replace_settings(GameSettings.create_finkel())

    def masters(self) -> 'GameBuilder':
        """
        Creates a new builder that allows the construction of games
        following the rules proposed by James Masters.
        """
        return self.replace_settings(GameSettings.create_masters())

    def aseb(self) -> 'GameBuilder':
        """
        Creates a new builder that allows the construction of games
        following the Aseb rules.
        """
        return self.replace_settings(GameSettings.create_aseb())

    def board_shape(self, board_shape: BoardShape | BoardType) -> 'GameBuilder':
        """
        Copies this game builder with the shape of the board updated.
        """
        return self.replace_settings(self._settings.with_board_shape(board_shape))

    def paths(self, paths: PathPair | PathType) -> 'GameBuilder':
        """
        Copies this game builder with the paths taken by each player updated.
        """
        return self.replace_settings(self._settings.with_paths(paths))

    def dice(self, dice: Callable[[], Dice] | DiceType) -> 'GameBuilder':
        """
        Copies this game builder with the factory used to generate dice updated.
        """
        return self.replace_settings(self._settings.with_dice(dice))

    def starting_piece_count(self, starting_piece_count: int) -> 'GameBuilder':
        """
        Copies this game builder with the number of starting
        pieces of each player updated.
        """
        return self.replace_settings(
            self._settings.with_starting_piece_count(starting_piece_count)
        )

    def safe_rosettes(self, safe_rosettes: bool) -> 'GameBuilder':
        """
        Copies this game builder with whether rosettes are safe
        from capture set to the given value.
        """
        return self.replace_settings(
            self._settings.with_safe_rosettes(safe_rosettes)
        )

    def rosettes_grant_extra_rolls(self, rosettes_grant_extra_rolls: bool) -> 'GameBuilder':
        """
        Copies this game builder with whether landing on a rosette
        grants an additional roll set to the given value.
        """
        return self.replace_settings(
            self._settings.with_rosettes_grant_extra_rolls(rosettes_grant_extra_rolls)
        )

    def captures_grant_extra_rolls(self, captures_grant_extra_rolls: bool) -> 'GameBuilder':
        """
        Copies this game builder with whether capturing a piece
        grants an additional roll set to the given value.
        """
        return self.replace_settings(
            self._settings.with_captures_grant_extra_rolls(captures_grant_extra_rolls)
        )

    def build_rules(self) -> RuleSet:
        """
        Generates a rule set to match the settings in this builder.
        """
        return self._rule_set_provider.create(self._settings, GameMetadata())

    def build(self) -> Game:
        """
        Generates a new game using the rules set in this builder.
        """
        return Game(self.build_rules())



from .tile import Tile
from .player import PlayerType
from enum import Enum
from typing import Callable


class PathPair:
    """
    Represents a pair of paths for the light and dark players to
    move their pieces along in a game of the Royal Game of Ur.
    """
    __slots__ = ("_name", "_light_with_ends", "_dark_with_ends", "_light", "_dark",)

    _name: str
    _light_with_ends: list[Tile]
    _dark_with_ends: list[Tile]
    _light: list[Tile]
    _dark: list[Tile]

    def __init__(
            self,
            name: str,
            lightWithStartEnd: list[Tile],
            darkWithStartEnd: list[Tile],
    ):
        self._name = name
        self._light_with_ends = [*lightWithStartEnd]
        self._dark_with_ends = [*darkWithStartEnd]
        self._light = self._light_with_ends[1:-1]
        self._dark = self._dark_with_ends[1:-1]

    @property
    def name(self) -> str:
        """
        The name of this path pair.
        """
        return self.name

    @property
    def light_with_ends(self) -> list[Tile]:
        """
        The path that light players take around the board, including
        the start and end tiles that exist off the board.
        """
        return self._light_with_ends

    @property
    def dark_with_ends(self) -> list[Tile]:
        """
        The path that dark players take around the board, including
        the start and end tiles that exist off the board.
        """
        return self._dark_with_ends

    @property
    def light(self) -> list[Tile]:
        """
        The path that light players take around the board, excluding
        the start and end tiles that exist off the board.
        """
        return self._light

    @property
    def dark(self) -> list[Tile]:
        """
        The path that dark players take around the board, excluding
        the start and end tiles that exist off the board.
        """
        return self._dark

    @property
    def light_start(self) -> Tile:
        """
        The start tile of the light player that exists off the board.
        """
        return self._light_with_ends[0]

    @property
    def light_end(self) -> Tile:
        """
        The end tile of the light player that exists off the board.
        """
        return self._light_with_ends[-1]

    @property
    def dark_start(self) -> Tile:
        """
        The start tile of the dark player that exists off the board.
        """
        return self._dark_with_ends[0]

    @property
    def dark_end(self) -> Tile:
        """
        The end tile of the dark player that exists off the board.
        """
        return self._dark_with_ends[-1]

    def get(self, player: PlayerType) -> list[Tile]:
        """
        Gets the path of the given player, excluding the start and
        end tiles that exist off the board.
        """
        if player == PlayerType.LIGHT:
            return self._light
        elif player == PlayerType.DARK:
            return self._dark
        else:
            raise ValueError(f"Unknown PlayerType {player}")

    def get_with_ends(self, player: PlayerType) -> list[Tile]:
        """
        Gets the path of the given player, including the start and
        end tiles that exist off the board.
        """
        if player == PlayerType.LIGHT:
            return self._light_with_ends
        elif player == PlayerType.DARK:
            return self._dark_with_ends
        else:
            raise ValueError(f"Unknown PlayerType {player}")

    def get_start(self, player: PlayerType) -> list[Tile]:
        """
        Gets the start tile of the given player, which exists off the board.
        """
        return self.get_with_ends(player)[0]

    def get_end(self, player: PlayerType) -> list[Tile]:
        """
        Gets the end tile of the given player, which exists off the board.
        """
        return self.get_with_ends(player)[-1]

    def is_equivalent(self, other: 'PathPair') -> bool:
        """
        Determines whether this set of paths and other cover the same tiles,
        in the same order. This does not check the start and end tiles that
        exist off the board, or the name of the paths.
        """
        return self._light == other._light

    def __eq__(self, other: object) -> bool:
        if type(other) is not type(self):
            return False

        return self._name == other._name \
            and self._light_with_ends == other._light_with_ends \
            and self._dark_with_ends == other._dark_with_ends


class AsebPathPair(PathPair):
    """
    The standard paths that are used for Aseb.

    Citation: W. Crist, A.E. Dunn-Vaturi, and A. de Voogt,
    Ancient Egyptians at Play: Board Games Across Borders,
    Bloomsbury Egyptology, Bloomsbury Academic, London, 2016.
    """
    __slots__ = ()

    NAME: str = "Aseb"
    """
    The name of this type of path pair.
    """

    LIGHT_PATH: list[Tile] = Tile.create_path(
        (1, 5),
        (1, 1),
        (2, 1),
        (2, 12),
        (1, 12),
    )
    """
    The path of the light player's pieces.
    """

    DARK_PATH: list[Tile] = Tile.create_path(
        (3, 5),
        (3, 1),
        (2, 1),
        (2, 12),
        (3, 12),
    )
    """
    The path of the dark player's pieces.
    """

    def __init__(self):
        super().__init__(
            AsebPathPair.NAME,
            AsebPathPair.LIGHT_PATH,
            AsebPathPair.DARK_PATH
        )


class BellPathPair(PathPair):
    """
    The paths proposed by Bell for the Royal Game of Ur.

    Citation: R.C. Bell, Board and Table Games From Many Civilizations,
    revised ed., Vol. 1 and 2, Dover Publications, Inc., New York, 1979.
    """
    __slots__ = ()

    NAME: str = "Bell"
    """
    The name of this type of path pair.
    """

    LIGHT_PATH: list[Tile] = Tile.create_path(
        (1, 5),
        (1, 1),
        (2, 1),
        (2, 8),
        (1, 8),
        (1, 6),
    )
    """
    The path of the light player's pieces.
    """

    DARK_PATH: list[Tile] = Tile.create_path(
        (3, 5),
        (3, 1),
        (2, 1),
        (2, 8),
        (3, 8),
        (3, 6),
    )
    """
    The path of the dark player's pieces.
    """

    def __init__(self):
        super().__init__(
            BellPathPair.NAME,
            BellPathPair.LIGHT_PATH,
            BellPathPair.DARK_PATH
        )


class MastersPathPair(PathPair):
    """
    The paths proposed by Masters for the Royal Game of Ur.

    Citation: J. Masters, The Royal Game of Ur &amp; The Game of 20 Squares (2021).
    Available at https://www.tradgames.org.uk/games/Royal-Game-Ur.htm.
    """
    __slots__ = ()

    NAME: str = "Masters"
    """
    The name of this type of path pair.
    """

    LIGHT_PATH: list[Tile] = Tile.create_path(
        (1, 5),
        (1, 1),
        (2, 1),
        (2, 7),
        (3, 7),
        (3, 8),
        (1, 8),
        (1, 6),
    )
    """
    The path of the light player's pieces.
    """

    DARK_PATH: list[Tile] = Tile.create_path(
        (3, 5),
        (3, 1),
        (2, 1),
        (2, 7),
        (1, 7),
        (1, 8),
        (3, 8),
        (3, 6),
    )
    """
    The path of the dark player's pieces.
    """

    def __init__(self):
        super().__init__(
            MastersPathPair.NAME,
            MastersPathPair.LIGHT_PATH,
            MastersPathPair.DARK_PATH
        )


class MurrayPathPair(PathPair):
    """
    The paths proposed by Murray for the Royal Game of Ur.

    Citation: H.J.R. Murray, A History of Board-games Other Than Chess,
    Oxford University Press, Oxford, 1952.
    """
    __slots__ = ()

    NAME: str = "Murray"
    """
    The name of this type of path pair.
    """

    LIGHT_PATH: list[Tile] = Tile.create_path(
        (1, 5),
        (1, 1),
        (2, 1),
        (2, 7),
        (3, 7),
        (3, 8),
        (1, 8),
        (1, 7),
        (2, 7),
        (2, 1),
        (3, 1),
        (3, 5),
    )
    """
    The path of the light player's pieces.
    """

    DARK_PATH: list[Tile] = Tile.create_path(
        (3, 5),
        (3, 1),
        (2, 1),
        (2, 7),
        (1, 7),
        (1, 8),
        (3, 8),
        (3, 7),
        (2, 7),
        (2, 1),
        (1, 1),
        (1, 5),
    )
    """
    The path of the dark player's pieces.
    """

    def __init__(self):
        super().__init__(
            MurrayPathPair.NAME,
            MurrayPathPair.LIGHT_PATH,
            MurrayPathPair.DARK_PATH
        )


class SkiriukPathPair(PathPair):
    """
    The paths proposed by Skiriuk for the Royal Game of Ur.

    Citation: D. Skiriuk, The rules of royal game of ur (2021).
    Available at https://skyruk.livejournal.com/231444.html.
    """
    __slots__ = ()

    NAME: str = "Skiriuk"
    """
    The name of this type of path pair.
    """

    LIGHT_PATH: list[Tile] = Tile.create_path(
        (1, 5),
        (1, 1),
        (2, 1),
        (2, 7),
        (3, 7),
        (3, 8),
        (1, 8),
        (1, 7),
        (2, 7),
        (2, 0),
    )
    """
    The path of the light player's pieces.
    """

    DARK_PATH: list[Tile] = Tile.create_path(
        (3, 5),
        (3, 1),
        (2, 1),
        (2, 7),
        (1, 7),
        (1, 8),
        (3, 8),
        (3, 7),
        (2, 7),
        (2, 0),
    )
    """
    The path of the dark player's pieces.
    """

    def __init__(self):
        super().__init__(SkiriukPathPair.NAME, SkiriukPathPair.LIGHT_PATH, SkiriukPathPair.DARK_PATH)


class PathType(Enum):
    """
    The type of path to use in a game.
    """

    BELL = (1, BellPathPair.NAME, lambda: BellPathPair())
    """
    The path proposed by Bell for the Royal Game of Ur.
    """

    ASEB = (2, AsebPathPair.NAME, lambda: AsebPathPair())
    """
    The standard path used for Aseb.
    """

    MASTERS = (3, MastersPathPair.NAME, lambda: MastersPathPair())
    """
    The path proposed by Masters for the Royal Game of Ur.
    """

    MURRAY = (4, MurrayPathPair.NAME, lambda: MurrayPathPair())
    """
    The path proposed by Murray for the Royal Game of Ur.
    """

    SKIRIUK = (5, SkiriukPathPair.NAME, lambda: SkiriukPathPair())
    """
    The path proposed by Skiriuk for the Royal Game of Ur.
    """

    def __init__(self, value: int, text_name: str, create_path_pair: Callable[[], PathPair]):
        self._value_ = value
        self._text_name = text_name
        self._create_path_pair = create_path_pair

    @property
    def text_name(self) -> str:
        """
        The name of these paths.
        """
        return self._text_name

    def create_path_pair(self) -> PathPair:
        """
        Create an instance of the paths.
        """
        return self._create_path_pair()



import random
from enum import Enum
from abc import ABC, abstractmethod
from overrides import overrides
from typing import Callable


class Roll:
    """
    A roll of the dice.
    """

    __slots__ = ("_value",)

    _value: int

    def __init__(self, value: int):
        self._value = value

    @property
    def value(self) -> int:
        """
        The value of the dice that was rolled.
        """
        return self._value


class Dice(ABC):
    """
    A generator of dice rolls.
    """
    __slots__ = ("_name",)

    _name: str

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        """
        The name of this dice.
        """
        return self._name

    def has_state() -> bool:
        """
        Returns whether this dice holds any state that affects its dice rolls.
        If this is overriden, then copy_from should also be overriden.
        """
        return False

    def copy_from(self, other: 'Dice') -> bool:
        """
        Copies the state of the other dice into this dice. If the dice does
        not have state, this is a no-op.
        """
        pass

    @abstractmethod
    def get_max_roll_value(self) -> int:
        """
        Gets the maximum value that could be rolled by this dice.
        """
        pass

    @abstractmethod
    def get_roll_probabilities(self) -> list[float]:
        """
        Gets the probability of rolling each value of the dice, where the
        index into the returned array represents the value of the roll.
        """
        pass

    @abstractmethod
    def roll_value(self) -> int:
        """
        Generates a random roll using this dice, and returns just the value.
        If this dice has state, this should call record_roll.
        """
        pass

    def record_roll(self, value: int):
        """
        Updates the state of this dice after having rolled value.
        """
        pass

    @abstractmethod
    def generate_roll(self, value: int) -> Roll:
        """
        Generates a roll with the given value using this dice.
        """
        pass

    def roll(self, value: int | None = None) -> Roll:
        """
        Generates a roll using this dice. If no value is provided, then the
        value of the roll will be random.
        """
        if value is None:
            value = self.roll_value()
        else:
            self.record_roll(value)

        return self.generate_roll(value)


class BinaryDice(Dice):
    """
    Rolls a number of binary die and counts the result.
    """
    __slots__ = ("_num_die", "_roll_probabilities")

    _num_die: int
    _roll_probabilities: list[float]

    def __init__(self, name: str, num_die: int):
        super().__init__(name)
        self._num_die = num_die
        self._roll_probabilities = []

        # Binomial Distribution
        baseProb = 0.5 ** num_die
        nChooseK = 1
        for roll in range(0, num_die + 1):
            self._roll_probabilities.append(baseProb * nChooseK)
            nChooseK = nChooseK * (num_die - roll) // (roll + 1)

    @property
    def num_die(self) -> int:
        """
        The number of binary dice to roll.
        """
        return self._num_die

    @overrides
    def get_max_roll_value(self) -> int:
        return self._num_die

    @overrides
    def get_roll_probabilities(self) -> list[float]:
        return self._roll_probabilities

    @overrides
    def roll_value(self) -> int:
        value = 0
        for _ in range(self._num_die):
            # Simulate rolling a single D2 dice.
            if random.random() >= 0.5:
                value += 1

        return value

    @overrides
    def generate_roll(self, value: int) -> Roll:
        if value < 0 or value > self._num_die:
            raise ValueError(f"This dice cannot roll {value}");

        return Roll(value)


class BinaryDice0AsMax(BinaryDice):
    """
    A set of binary dice where a roll of zero actually represents
    the highest roll possible, rather than the lowest.
    """
    __slots__ = ("_max_roll_value",)

    _max_roll_value: int

    def __init__(self, name: str, num_die: int):
        super().__init__(name, num_die)
        self._max_roll_value = num_die + 1
        self._roll_probabilities = [
            0,
            *self._roll_probabilities[1:],
            self._roll_probabilities[0]
        ]

    @overrides
    def get_max_roll_value(self) -> int:
        return self.max_roll_value

    @overrides
    def roll_value(self) -> int:
        value = super().roll_value()
        return self.max_roll_value if value == 0 else value

    @overrides
    def generate_roll(self, value: int) -> Roll:
        if value <= 0 or value > self.max_roll_value:
            raise ValueError(f"This dice cannot roll {value}");

        return Roll(value)


class DiceType(Enum):
    """
    The type of dice to use in a game.
    """

    FOUR_BINARY = (1, "FourBinary", lambda: BinaryDice("FourBinary", 4))
    """
    The standard board shape.
    """

    THREE_BINARY_0MAX = (2, "ThreeBinary0Max", lambda: BinaryDice0AsMax("ThreeBinary0Max", 3))
    """
    The Aseb board shape.
    """

    def __init__(self, value: int, text_name: str, create_dice: Callable[[], Dice]):
        self._value_ = value
        self._text_name = text_name
        self._create_dice = create_dice

    @property
    def text_name(self) -> str:
        """
        The name of this dice.
        """
        return self._text_name

    @property
    def dice_factory(self) -> Callable[[], Dice]:
        """
        The factory that is used to produce this type of dice.
        """
        return self._create_dice

    def create_dice(self) -> Dice:
        """
        Creates a set of these dice.
        """
        return self._create_dice()
